/*
 * Copyright (C) 2019  Kevin Whitaker <eyecreate@eyecreate.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "miflora.h"
#include <QDataStream>

void MiFlora::startSearch()
{
    qDebug() << "Starting search...";
    agent = new QBluetoothDeviceDiscoveryAgent();
    agent->setLowEnergyDiscoveryTimeout(5000);
    devices.clear();
    connect(agent, &QBluetoothDeviceDiscoveryAgent::deviceDiscovered, this, &MiFlora::foundDevice);
    agent->start(QBluetoothDeviceDiscoveryAgent::LowEnergyMethod);
}

void MiFlora::stopSearch()
{
    if(agent != NULL) { 
        qDebug() << "Stopping Search.";
        agent->stop();
        agent->disconnect();
        delete agent;
    }
}

void MiFlora::updateDataFromDevice ( QString mac )
{
    for(BluetoothDevices *info: devices) {
        if(info->device().address().toString() == mac) {
            qDebug() << "Discovering on:" + info->device().name();
            currentController = QLowEnergyController::createCentral(info->device());
            currentController->setRemoteAddressType(QLowEnergyController::PublicAddress);
            connect(currentController, &QLowEnergyController::serviceDiscovered, this, &MiFlora::serviceDiscovered);
            connect(currentController, static_cast<void (QLowEnergyController::*)(QLowEnergyController::Error)>(&QLowEnergyController::error), this, &MiFlora::logControllerError);
            connect(currentController, &QLowEnergyController::connected, this, [this]{
                currentController->discoverServices();
            });
            currentController->connectToDevice();
        }
    }
}

void MiFlora::foundDevice ( const QBluetoothDeviceInfo& device )
{
    if(device.coreConfigurations() & QBluetoothDeviceInfo::LowEnergyCoreConfiguration && device.address().toString().startsWith("C4:7C")) {
        devices.append(new BluetoothDevices(device));
        qDebug() << "Added device";
        emit newDeviceFound();
    }
}

QQmlListProperty<BluetoothDevices> MiFlora::getDeviceList()
{
    return QQmlListProperty<BluetoothDevices>(this, devices);
}

void MiFlora::logControllerError ( QLowEnergyController::Error err )
{
    qDebug() << "Error:" << err;
}

void MiFlora::logServiceError(QLowEnergyService::ServiceError err)
{
    qDebug() << "Service Error:" << err;
}


void MiFlora::serviceStateChanges ( QLowEnergyService::ServiceState state )
{
    if(state == QLowEnergyService::ServiceState::ServiceDiscovered) {
        for(QLowEnergyCharacteristic chrt : sensorService->characteristics()) {
            if(chrt.uuid().toUInt16() == magicChar) {
                sensorService->writeCharacteristic(chrt, QByteArray::fromHex("a01f"));
            }
        }
    }
}

void MiFlora::serviceCharWritten(QLowEnergyCharacteristic changedChar, QByteArray value)
{
    if(changedChar.uuid().toUInt16() == magicChar) {
        //Used to only check sensor data after writting magic bytes.
        for(QLowEnergyCharacteristic chrt : sensorService->characteristics()) {
            if(chrt.uuid().toUInt16() == batteryFirmwareChar) {
                sensorService->readCharacteristic(chrt);
            }
            if(chrt.uuid().toUInt16() == sensorsChar) {
                sensorService->readCharacteristic(chrt);
            }
        }
    }
}

void MiFlora::serviceCharRead(QLowEnergyCharacteristic readChar, QByteArray value)
{
    if(readChar.uuid().toUInt16() == batteryFirmwareChar) {
        QDataStream parser(value);
        parser >> battery;
        parser.skipRawData(1);
        QByteArray version_buf(5, Qt::Uninitialized);
        parser.readRawData(version_buf.data(), 5);
        version = QString(version_buf);
        emit batteryChanged(battery);
        qDebug() << "Firmware: " << version;
    } else if(readChar.uuid().toUInt16() == sensorsChar) {
        quint16 origTemp;
        QDataStream parser(value);
        parser.setByteOrder(QDataStream::LittleEndian);
        parser >> origTemp;
        temp = origTemp/ 10.0; //original value in 0.1 C
        parser.skipRawData(1);
        parser >> bright;
        parser >> moisture;
        parser >> conduct;
        emit brightnessChanged(bright);
        emit temperatureChanged(temp);
        emit moistureChanged(moisture);
        emit conductionChanged(conduct);
    }
}

void MiFlora::serviceDiscovered ( const QBluetoothUuid& gatt )
{
    if(gatt == QBluetoothUuid(dataService)) {
        sensorService = currentController->createServiceObject(gatt);
        connect(sensorService, &QLowEnergyService::stateChanged, this, &MiFlora::serviceStateChanges);
        connect(sensorService, static_cast<void (QLowEnergyService::*)(QLowEnergyService::ServiceError)>(&QLowEnergyService::error), this, &MiFlora::logServiceError);
        connect(sensorService, &QLowEnergyService::characteristicWritten, this, &MiFlora::serviceCharWritten);
        connect(sensorService, &QLowEnergyService::characteristicRead, this, &MiFlora::serviceCharRead);
        sensorService->discoverDetails();
    }
}



